from typing import Any, Optional, Union

from pydantic import BaseModel
from pydantic.fields import Field
from pydantic import root_validator


class Payload(BaseModel):
    def payload(self, token: str) -> dict[str, Union[str, int]]:
        payload_data = self.dict(by_alias=True, exclude_none=True)
        payload_data.update(token=token)
        return payload_data


class SearchPayload(Payload):
    name: str = ""
    author: str = ""
    difficulty_min: int = Field(default=1, ge=1, le=6, alias="difficulty-min")
    difficulty_max: int = Field(default=6, ge=1, le=6, alias="difficulty-max")
    quality_min: int = Field(default=1, ge=1, le=6, alias="quality-min")
    quality_max: int = Field(default=6, ge=1, le=6, alias="quality-max")
    lang: Optional[str] = None
    arch: Optional[str] = None
    platform: Optional[str] = None

    class Config:
        allow_population_by_field_name = True

    @root_validator
    def min_is_less_than_max(cls, values: dict) -> dict[str, Any]:  # noqa: WPS110
        min_difficulty: int = values["difficulty_min"]
        max_difficulty: int = values["difficulty_max"]
        if min_difficulty >= max_difficulty:
            raise ValueError("difficulty_min should be minimum be 1 less than difficulty_max")
        return values


class AuthPayload(Payload):
    username: str = Field(..., alias="name")
    password: str



import itertools
from pathlib import Path
from typing import Iterable

from requests.sessions import Session
import typer

from crackmes_dl.endpoints import CrackmeEndpoint
from crackmes_dl.endpoints import LastsEndpoint
from crackmes_dl.endpoints import LoginEndpoint
from crackmes_dl.endpoints import SearchEndpoint
from crackmes_dl.payloads import AuthPayload
from crackmes_dl.payloads import SearchPayload
from crackmes_dl.responses import CrackmeEntry
from crackmes_dl.responses import Link


class CrackmesApi:
    def __init__(self, domain: str) -> None:
        self._session = Session()
        self._login = LoginEndpoint(domain=domain)
        self._search = SearchEndpoint(domain=domain)
        self._crackme = CrackmeEndpoint(domain=domain)
        self._lasts = LastsEndpoint(domain=domain)

    def login(self, payload: AuthPayload) -> bool:
        return self._login.authenticate(session=self._session, payload=payload)

    def search(self, payload: SearchPayload) -> list[CrackmeEntry]:
        return self._search.search(session=self._session, payload=payload)

    def download_single(self, output_dir: Path, crackme_id: str) -> None:
        crackme = Link(text=crackme_id, url=f"crackme/{crackme_id}")
        self._crackme.download(
            session=self._session,
            crackme=crackme,
            output_dir=output_dir,
        )

    def download(self, output_dir: Path, crackmes: list[CrackmeEntry]) -> None:
        count = len(crackmes)
        with typer.progressbar(crackmes, label=f"Downloading {count} crackmes") as progress:
            for crackme in progress:
                self._crackme.download(
                    session=self._session,
                    crackme=crackme.crackme,
                    output_dir=output_dir,
                )

    def unlimited_search(self, search_terms: SearchPayload) -> Iterable[CrackmeEntry]:
        for page in itertools.count(start=1):
            typer.echo(f"Searching page {page}")
            crackmes_on_page = self.lasts(page=page)
            for crackme in crackmes_on_page:
                if crackme.matches_search_terms(search_terms=search_terms):
                    yield crackme
            if not crackmes_on_page:
                return

    def lasts(self, page: int) -> list[CrackmeEntry]:
        return self._lasts.get_crackmes_on_page(session=self._session, page=page)



from enum import Enum
from itertools import count
from pathlib import Path
from typing import Optional

import typer

from crackmes_dl.api import CrackmesApi
from crackmes_dl.payloads import SearchPayload

app = typer.Typer()

PasswordPrompt: str = typer.Option(..., prompt=True, hide_input=True)
RequiredPath: Path = typer.Option(...)
Id: str = typer.Option(..., help="fx: 617965c733c5d4329c345330")
Domain: str = typer.Option("https://crackmes.one")


class Languages(str, Enum):
    c_cpp = "C/C++"
    asm = "Assembler"
    java = "Java"
    vb = "(Visual) Basic"
    bd = "Borland Delphi"
    tp = "Turbo Pascal"
    net = ".NET"
    other = "Unspecified/other"


class Architectures(str, Enum):
    x86 = "x86"
    x86_64 = "X86-64"
    java = "java"
    arm = "ARM"
    mips = "MIPS"
    other = "other"


class Platforms(str, Enum):
    dos = "DOS"
    mac = "Mac OS X"
    multi = "Multiplatform"
    nix = "Unix/linux etc."
    win = "Windows"
    win2000_xp = "Windows 2000/XP only"
    win7 = "Windows 7 Only"
    win_vista = "Windows Vista Only"
    other = "Unspecificed/other"


Name: str = typer.Option("", help="Name of the crackme must include 'search string'")
Author: str = typer.Option("", help="Author name must include 'search string'")
Difficulty_min: int = typer.Option(1, help="Difficulty greater or equal to.")
Difficulty_max: int = typer.Option(6, help="Difficulty less or equal to.")
Quality_min: int = typer.Option(1, help="Quality greater or equal to.")
Quality_max: int = typer.Option(6, help="Quality less or equal to.")
Lang: Optional[Languages] = typer.Option(None, help="Defaults to including all")
Arch: Optional[Architectures] = typer.Option(None, help="Defaults to including all")
Platform: Optional[Platforms] = typer.Option(None, help="Defaults to including all")
Quick: bool = typer.Option(True, help="Faster but limited to max 50 results")  # noqa: WPS425


@app.command()
def download_all(
    output_dir: Path = RequiredPath,
    domain: str = Domain,
    starting_page: int = 1,
) -> None:
    api = CrackmesApi(domain=domain)
    for page in count(start=starting_page):
        typer.echo(f"Downloading page: {page}")
        crackmes = api.lasts(page=page)
        if not crackmes:
            break
        api.download(output_dir=output_dir, crackmes=crackmes)


@app.command()
def download(output_dir: Path = RequiredPath, crackme: str = Id, domain: str = "https://crackmes.one") -> None:
    api = CrackmesApi(domain=domain)
    api.download_single(crackme_id=crackme, output_dir=output_dir)


@app.command()
def search_and_download(  # noqa: WPS211
    output_dir: Path = RequiredPath,
    domain: str = Domain,
    quick: bool = Quick,
    name: str = Name,
    author: str = Author,
    difficulty_min: int = Difficulty_min,
    difficulty_max: int = Difficulty_max,
    quality_min: int = Quality_min,
    quality_max: int = Quality_max,
    lang: Optional[Languages] = Lang,
    arch: Optional[Architectures] = Arch,
    platform: Optional[Platforms] = Platform,
) -> None:
    api = CrackmesApi(domain=domain)
    search_terms = SearchPayload(
        name=name,
        author=author,
        difficulty_min=difficulty_min,
        difficulty_max=difficulty_max,
        quality_min=quality_min,
        quality_max=quality_max,
        lang=lang,
        arch=arch,
        platform=platform,
    )
    if quick:
        crackmes = api.search(payload=search_terms)
        api.download(output_dir=output_dir, crackmes=crackmes)
    else:
        for crackme in api.unlimited_search(search_terms=search_terms):
            api.download(output_dir=output_dir, crackmes=[crackme])


if __name__ == "__main__":
    app()



from datetime import datetime

from bs4.element import Tag
from pydantic import BaseModel

from crackmes_dl.payloads import SearchPayload


class Link(BaseModel):
    text: str
    url: str

    @classmethod
    def from_col(cls, col: Tag) -> "Link":
        anchor_tag = col.find_all(href=True)[0]
        return cls(text=col.text.strip(), url=anchor_tag["href"])


class CrackmeEntry(BaseModel):
    crackme: Link
    author: Link
    language: str
    arch: str
    difficulty: float
    quality: float
    platform: str
    uploaded_ts: datetime
    solutions_count: int
    comments_count: int

    @classmethod
    def from_row(cls, row: Tag) -> "CrackmeEntry":
        cols: list[Tag] = row.find_all(name="td")
        crackme = Link.from_col(cols[0])
        author = Link.from_col(cols[1])
        stripped: list[str] = [col.text.strip() for col in cols]
        return cls(
            crackme=crackme,
            author=author,
            language=stripped[2],
            arch=stripped[3],
            difficulty=float(stripped[4]),
            quality=float(stripped[5]),
            platform=stripped[6],
            uploaded_ts=datetime.strptime(stripped[7], "%I:%M %p %m/%d/%Y"),
            solutions_count=int(stripped[8]),
            comments_count=int(stripped[9]),
        )

    def matches_search_terms(self, search_terms: SearchPayload) -> bool:
        return all(
            [
                search_terms.name in self.crackme.text,
                search_terms.author in self.author.text,
                self.quality > search_terms.quality_min,
                self.quality < search_terms.quality_max,
                self.difficulty > search_terms.difficulty_min,
                self.difficulty < search_terms.difficulty_max,
                search_terms.lang is None or search_terms.lang == self.language,
                search_terms.platform is None or search_terms.platform == self.platform,
            ],
        )



from datetime import datetime
from io import BytesIO
import os
from pathlib import Path
from zipfile import ZipFile

from bs4 import BeautifulSoup
from bs4.element import ResultSet
from pydantic.main import BaseModel
from requests.models import Response
from requests.sessions import Session

from crackmes_dl.payloads import AuthPayload
from crackmes_dl.payloads import Payload
from crackmes_dl.payloads import SearchPayload
from crackmes_dl.responses import CrackmeEntry
from crackmes_dl.responses import Link

HTML_PARSER = "html.parser"


class Endpoint:
    def __init__(self, domain: str, endpoint: str) -> None:
        self.endpoint = domain + endpoint


class FormEndpoint(Endpoint):
    def post(self, session: Session, payload: Payload) -> Response:
        token: str = self._find_input_token(session=session)
        payload_data = payload.payload(token=token)
        return session.post(url=self.endpoint, data=payload_data)

    def _find_input_token(self, session: Session) -> str:
        response: Response = session.get(url=self.endpoint)
        soup = BeautifulSoup(markup=response.text, features=HTML_PARSER)
        found_input_tokens: ResultSet = soup.find_all(name="input", attrs={"id": "token"})
        if found_input_tokens:
            return found_input_tokens[0].get("value")
        raise Exception("Couldn't find input token")


class LoginEndpoint(FormEndpoint):
    def __init__(self, domain: str) -> None:
        super().__init__(domain=domain, endpoint="/login")

    def authenticate(self, session: Session, payload: AuthPayload) -> bool:
        login_response = self.post(session=session, payload=payload)
        login_response.raise_for_status()
        return "Login successful!" in login_response.text


class SearchEndpoint(FormEndpoint):
    def __init__(self, domain: str) -> None:
        super().__init__(domain=domain, endpoint="/search")

    def search(self, session: Session, payload: SearchPayload) -> list[CrackmeEntry]:
        search_response = self.post(session=session, payload=payload)
        search_response.raise_for_status()
        html = search_response.text
        return self._search_results(html=html)

    def _search_results(self, html: str) -> list[CrackmeEntry]:
        soup = BeautifulSoup(markup=html, features=HTML_PARSER)
        tables: ResultSet = soup.find_all(name="table", attrs={"class": "table-striped"})
        if len(tables) != 1:
            raise Exception("We didn't find exactly 1 table!")
        table = tables[0]
        body = table.find(name="tbody")
        rows = body.find_all(name="tr")
        return [CrackmeEntry.from_row(row) for row in rows]


class Metadata(BaseModel):
    crackme_url: str
    download_url: str
    author: str
    description: str
    language: str
    arch: str
    difficulty: float
    quality: float
    platform: str
    uploaded_ts: datetime

    @classmethod
    def from_html(cls, html: str, crackme_id: str, domain: str) -> "Metadata":  # noqa: WPS210
        soup = BeautifulSoup(markup=html, features=HTML_PARSER)
        section = soup.find_all(name="div", attrs={"class": "panel-background"})[0]
        p_tags = [p_tag.text.strip() for p_tag in section.find_all(name="p")]
        author: str = p_tags[0].replace("Author: ", "")
        description: str = p_tags[8]
        language: str = p_tags[1].replace("Language: ", "")
        arch: str = p_tags[6].replace("Arch: ", "")
        difficulty: float = float(p_tags[4].replace("Difficulty: ", ""))
        quality: float = float(p_tags[5].replace("Quality: ", ""))
        platform: str = p_tags[3].replace("Platform\n", "").strip()
        uploaded: str = p_tags[2].replace("Upload: ", "").strip()
        uploaded_ts: datetime = datetime.strptime(uploaded, "%I:%M %p %m/%d/%Y")

        return Metadata(
            crackme_url=f"{domain}/crackme/{crackme_id}",
            download_url=f"{domain}/static/crackme/{crackme_id}.zip",
            author=author,
            description=description,
            language=language,
            arch=arch,
            difficulty=difficulty,
            quality=quality,
            platform=platform,
            uploaded_ts=uploaded_ts,
        )


class CrackmeEndpoint(Endpoint):
    def __init__(self, domain: str) -> None:
        self.domain = domain
        super().__init__(domain=domain, endpoint="/crackme")

    def download(self, session: Session, crackme: Link, output_dir: Path) -> None:  # noqa: WPS210
        crackme_id: str = crackme.url.split("/")[-1]
        output_path = output_dir.joinpath(crackme_id)
        if self._file_already_exists(output_path=output_path):
            return
        os.makedirs(output_path)
        metadata = self._find_metadata(session=session, crackme_id=crackme_id)
        crackme_dl_response: Response = session.get(metadata.download_url)
        encrypted_zip = ZipFile(BytesIO(crackme_dl_response.content))
        pwd = b"crackmes.de" if metadata.author == "crackmes.de" else b"crackmes.one"
        encrypted_zip.extractall(path=output_path, pwd=pwd)
        with open(file=output_path.joinpath("metadata.json"), mode="w+") as json_file:
            json_file.write(metadata.json())

    def _find_metadata(self, session: Session, crackme_id: str) -> Metadata:
        response: Response = session.get(f"{self.endpoint}/{crackme_id}")
        response.raise_for_status()
        return Metadata.from_html(html=response.text, crackme_id=crackme_id, domain=self.domain)

    def _file_already_exists(self, output_path: Path) -> bool:
        return output_path.is_file()


class LastsEndpoint(Endpoint):
    def __init__(self, domain: str) -> None:
        super().__init__(domain=domain, endpoint="/lasts")

    def get_crackmes_on_page(self, session: Session, page: int) -> list[CrackmeEntry]:
        response: Response = session.get(url=f"{self.endpoint}/{page}")
        soup = BeautifulSoup(markup=response.text, features=HTML_PARSER)
        rows = soup.find_all(name="tr", attrs={"class": "text-center"})
        return [CrackmeEntry.from_row(row) for row in rows]
